@@ -1,7 +1,6 @@
 use pyo3::types::*;
 use serde::de::{self, IntoDeserializer};
 use serde::Deserialize;
-use std::convert::TryInto;
 
 use crate::error::{PythonizeError, Result};
 
@@ -26,21 +25,17 @@ impl<'de> Depythonizer<'de> {
     fn sequence_access(&self, expected_len: Option<usize>) -> Result<PySequenceAccess<'de>> {
         let seq: &PySequence = self.input.downcast()?;
         let len = seq.len()?;
-        let len_usize: usize = len.try_into().expect("negative sequence length");
 
         match expected_len {
-            Some(expected) if expected != len_usize => Err(
-                PythonizeError::incorrect_sequence_length(expected, len_usize),
-            ),
+            Some(expected) if expected != len => {
+                Err(PythonizeError::incorrect_sequence_length(expected, len))
+            }
             _ => Ok(PySequenceAccess::new(seq, len)),
         }
     }
 
-    fn dict_access(
-        &self,
-    ) -> Result<PyDictAccess<'de, impl Iterator<Item = (&'de PyAny, &'de PyAny)>>> {
-        let dict: &PyDict = self.input.downcast()?;
-        Ok(PyDictAccess::new(dict.iter()))
+    fn dict_access(&self) -> Result<PyMappingAccess<'de>> {
+        PyMappingAccess::new(self.input.downcast()?)
     }
 }
 
@@ -90,6 +85,10 @@ impl<'a, 'de> de::Deserializer<'de> for &'a mut Depythonizer<'de> {
             self.deserialize_tuple(obj.len()?, visitor)
         } else if obj.is_instance::<PyUnicode>()? {
             self.deserialize_str(visitor)
+        } else if let Ok(seq) = obj.downcast::<PySequence>() {
+            self.deserialize_tuple(seq.len()?, visitor)
+        } else if obj.downcast::<PyMapping>().is_ok() {
+            self.deserialize_map(visitor)
         } else {
             Err(PythonizeError::unsupported_type(
                 obj.get_type().name().unwrap_or("<unknown>"),
@@ -256,7 +255,7 @@ impl<'a, 'de> de::Deserializer<'de> for &'a mut Depythonizer<'de> {
             }
             let variant: &PyString = d
                 .keys()
-                .get_item(0)
+                .get_item(0)?
                 .cast_as()
                 .map_err(|_| PythonizeError::dict_key_not_string())?;
             let value = d.get_item(variant).unwrap();
@@ -291,12 +290,12 @@ impl<'a, 'de> de::Deserializer<'de> for &'a mut Depythonizer<'de> {
 
 struct PySequenceAccess<'a> {
     seq: &'a PySequence,
-    index: isize,
-    len: isize,
+    index: usize,
+    len: usize,
 }
 
 impl<'a> PySequenceAccess<'a> {
-    fn new(seq: &'a PySequence, len: isize) -> Self {
+    fn new(seq: &'a PySequence, len: usize) -> Self {
         Self { seq, index: 0, len }
     }
 }
@@ -318,43 +317,42 @@ impl<'de> de::SeqAccess<'de> for PySequenceAccess<'de> {
     }
 }
 
-struct PyDictAccess<'de, Iter>
-where
-    Iter: Iterator<Item = (&'de PyAny, &'de PyAny)> + 'de,
-{
-    iter: Iter, // TODO: figure out why PyDictIterator is not publicly accessible upstream?
-    next_value: Option<&'de PyAny>,
+struct PyMappingAccess<'de> {
+    keys: &'de PySequence,
+    values: &'de PySequence,
+    key_idx: usize,
+    val_idx: usize,
+    len: usize,
 }
 
-impl<'de, Iter> PyDictAccess<'de, Iter>
-where
-    Iter: Iterator<Item = (&'de PyAny, &'de PyAny)> + 'de,
-{
-    fn new(iter: Iter) -> Self {
-        Self {
-            iter,
-            next_value: None,
-        }
+impl<'de> PyMappingAccess<'de> {
+    fn new(map: &'de PyMapping) -> Result<Self> {
+        let keys = map.keys()?;
+        let values = map.values()?;
+        let len = map.len()?;
+        Ok(Self {
+            keys,
+            values,
+            key_idx: 0,
+            val_idx: 0,
+            len,
+        })
     }
 }
 
-impl<'de, Iter> de::MapAccess<'de> for PyDictAccess<'de, Iter>
-where
-    Iter: Iterator<Item = (&'de PyAny, &'de PyAny)> + 'de,
-{
+impl<'de> de::MapAccess<'de> for PyMappingAccess<'de> {
     type Error = PythonizeError;
 
     fn next_key_seed<K>(&mut self, seed: K) -> Result<Option<K::Value>>
     where
         K: de::DeserializeSeed<'de>,
     {
-        match self.iter.next() {
-            None => Ok(None),
-            Some((key, value)) => {
-                self.next_value = Some(value);
-                seed.deserialize(&mut Depythonizer::from_object(key))
-                    .map(Some)
-            }
+        if self.key_idx < self.len {
+            let mut item_de = Depythonizer::from_object(self.keys.get_item(self.key_idx)?);
+            self.key_idx += 1;
+            seed.deserialize(&mut item_de).map(Some)
+        } else {
+            Ok(None)
         }
     }
 
@@ -362,11 +360,9 @@ where
     where
         V: de::DeserializeSeed<'de>,
     {
-        let value = self
-            .next_value
-            .take()
-            .expect("call next_value_seed after next_key_seed");
-        seed.deserialize(&mut Depythonizer::from_object(value))
+        let mut item_de = Depythonizer::from_object(self.values.get_item(self.val_idx)?);
+        self.val_idx += 1;
+        seed.deserialize(&mut item_de)
     }
 }
 
